// This demo is for testing only.
// Not to be distributed in the official release.
// cmdline.c
// Copyright (C) 2007-2009  Sensoray Co., Inc.  This program is free software;
// Linux Command line demo for 817.
// No X windows required. No openGL required
// For evaluation of the 817 without display.
// To determine FPS and 817 capabilities without the complications of
// displaying images.
// Demonstrates multithreaded usage of the 817 API/SDK
//
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You must obey the GNU General Public License in all respects for all of
// the code used. If you modify this file, you may extend this exception
// to your version of the file, but you are not obligated to do so. If you
// do not wish to do so, delete this exception statement from your version.

#define _GNU_SOURCE  // for O_DIRECT

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <pthread.h>
#include <sys/time.h>
#include <time.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include "s817.h"
#include "s817f.h"

#define TRUE                1
#define FALSE               0

#define MAXLEN              80         // Maximum length of user command string.

/* stream modes */
#define MODE_IDLE           0
#define MODE_STREAM         1


#define MAX_BOARDS (SYS_BOARDS)
static int G_numboards = 0; // number of boards detected 
#define MAX_CHANNELS SYS_CHAN_PER_BOARD

static int G_fwver = 0;
static int G_verbose = 0;
static int G_rsrv[MAX_CHANNELS];
static int jdec;
static int disk_mode=0;

static pthread_t G_io_thread_idJ; // Handle to jpeg processing thread.
static pthread_t G_io_thread_idB; // Handle to bmp processing thread.
static int    G_smode;        // Stream mode.
static int    G_stop_rq;      // Stream Halt requested.
static int    G_jpeg_load;
static int    G_jpegs_saved[MAX_CHANNELS];

typedef struct BOARD {                          // BOARD OBJECT -------
    int             smode;        // Stream mode.
    int             stop_rq;      // Stream Halt requested.
    char            fname[MAXLEN];              // File name for snapshots
    unsigned int    frame_stats_jpg[MAX_CHANNELS];  // Number of frames(used for FPS calc only)
    unsigned int    frame_stats_bmp[MAX_CHANNELS];  // 
    unsigned int jpg_frame_count[MAX_CHANNELS];
    unsigned int bmp_frame_count[MAX_CHANNELS];
    int snapshot_ppm[MAX_CHANNELS];
    int snapshot_jpeg[MAX_CHANNELS];
    MODE            mode[MAX_CHANNELS];         // 817 Mode structure
    int             board;                      // board number
    int             print_stats;
    int             b_stat_exclude[MAX_CHANNELS]; // don't update stats if just started


    pthread_t       io_thread_idM; // Handle to stream i/o thread.
} BOARD;

struct fifo_element {
    CAP_BUFFER cBuf;
    int        board;
};


static struct BOARD bd[SYS_BOARDS];     // Board objects.
static int act_bdnum   = 0;    // Board 0 is active by default.


// Function prototypes ////////////////////////////////////////////////////

static int app_init(void);
static void app_shutdown( void );
static void cmd_processor( void );

static void *main_thread_func( void* arg );

static void *jpg_thread_func(void *arg);
static void *bmp_thread_func(void *arg);

static void get_attributes( struct BOARD *b );
static void start_stream( struct BOARD *b, int chan, int capType );
static void stop_stream( struct BOARD *b, int chan );
static void stop_threads(void);
static int start_threads(void);
static void print_ecode(ECODE ecode);
static void getPrintErrorInfo(struct BOARD *b);
static void prompt( void );
static void print_stats(BOARD *b, double seconds);
static double current_time(void);
static void getSysStatus(struct BOARD *b);
static int exec_command( char *cmd, char *arg,char *arg2 );

static void querySysStatus(struct BOARD *b);


#define FIFO_NAME_JPG "s817_fifo_jpg_all"
#define FIFO_NAME_BMP "s817_fifo_bmp_all"

void *bmp_thread_func( void *arg )
{
    unsigned char          *rgb_frame; // for snapshots
    int rc;
    int board;
    int fifo_fd;
    CAP_STAT *capStat;
    int aChan;
    struct fifo_element fe;

    rgb_frame = (unsigned char *) malloc(768*576*3);

    if (rgb_frame == NULL) {
        printf("out of memory\n");
        return NULL;
    }
    fifo_fd = open(FIFO_NAME_BMP, O_RDONLY);
    if (fifo_fd == -1) {
        printf("could not open fifo\n");
        return NULL;
    }

    while (!G_stop_rq) {
        rc = read(fifo_fd, &fe, sizeof(struct fifo_element));
        if (rc <= 0)
            continue;

        capStat = (CAP_STAT *) fe.cBuf.pStat;
        board = fe.board;
        aChan = capStat->chanId;

        if (aChan < 1 || aChan > MAX_CHANNELS)
            continue;

        if (capStat->chanId != aChan)
            printf ("Chan ID error on chan %d: %d\n", aChan, capStat->chanId);
                
        G_rsrv[aChan-1] = capStat->reserved[1];
        // if processing BMPs, do it here 



        // release the buffer 
        rc = S817_ReleaseBuffer(board, aChan, fe.cBuf.bufId, CAP_BMP);
        if (rc != 0) {
            printf("failed to release buffer\n");
        }
    }
    close(fifo_fd);
    return NULL;
}


//jpeg thread func
void *jpg_thread_func( void *arg )
{
    CAP_STAT    *capStat;
    int rc;
    int board;
    int fifo_fd;
    int aChan;
    struct fifo_element fe;
    FILE *fout = NULL;
    int  fdout = -1;
    unsigned long wr_b;
    fifo_fd = open(FIFO_NAME_JPG, O_RDONLY);
    if (fifo_fd == -1) {
        printf("could not open fifo jpg\n");
        return NULL;
    }
    if (disk_mode) {
        fdout = open("jpeg_flat", O_WRONLY | O_DIRECT);
        if (fdout == -1) {
            printf("could not create file!\n");
            close(fifo_fd);
            return NULL;
        }

    } else {
        fout = fopen("jpeg_flat", "wb");
        if (fout == NULL) {
            printf("could not create file!\n");
            close(fifo_fd);
            return NULL;
        }
    }

    wr_b = 0;
    while (!G_stop_rq) {
        rc = read(fifo_fd, &fe, sizeof(struct fifo_element));
        if (rc <= 0) {
            continue;
        }
        capStat = (CAP_STAT *) fe.cBuf.pStat;
        board = fe.board;
        aChan = capStat->chanId;
        if (capStat->jpgsize && capStat->valid) {
            if (disk_mode) {
                write(fdout, fe.cBuf.pData, capStat->jpgsize);

            } else {
                fwrite(fe.cBuf.pData, 1, capStat->jpgsize, fout);
            }
            wr_b += capStat->jpgsize;
            G_jpegs_saved[aChan-1]++;
            // close file and restart after 900Mb
            if (!disk_mode) {
                if (wr_b > 900*1024*1024) {
                    printf("900Mb reached, fseek to start of file\n");
                    fseek(fout, SEEK_SET, 0);
                    wr_b = 0;
                }
            } else {
                if (wr_b > 900*1024*1024) {
                    printf("900Mb reached, starting file again\n");
                    fdout = open("jpeg_flat", O_WRONLY | O_DIRECT);
                    if (fdout == -1) {
                        printf("could not create file!\n");
                        close(fifo_fd);
                        return NULL;
                    }
                    wr_b = 0;
                }
            }
        }
        G_rsrv[aChan-1] = capStat->reserved[1];
        /* we are done with the buffer.  We must release it
           now so the driver can reuse it */
        rc = S817_ReleaseBuffer(board, aChan, fe.cBuf.bufId, CAP_JPG);
        if (rc != 0) {
            printf("release buffer JPG failed RC:%d"
                   "chan:%d bufID:%d\n", rc, aChan, fe.cBuf.bufId);
        }
    }
    if (disk_mode) {
        close(fdout);
    } else {
        fclose(fout);
    }
    close(fifo_fd);
    return NULL;
}

//  capture function (one per board)
void *main_thread_func( void *arg )
{
    BOARD                  *b;
    CAP_BUFFER  cBuf;
    ECODE ecode;
    int rc;
    int board;
    int fifo_fd1, fifo_fd2;
    int aChan, capType;
    double t = current_time();
    double tRot0 = -1.0, tRate0 = -1.0;
    double dt;
    struct fifo_element fe;
    b = (BOARD *) arg;
    board = b->board;
    fifo_fd1 = open(FIFO_NAME_BMP, O_WRONLY);
    if (fifo_fd1 == -1) {
        printf("could not open BMP fifo\n");
        return NULL;
    }
    fifo_fd2 = open(FIFO_NAME_JPG, O_WRONLY);
    if (fifo_fd2 == -1) {
        printf("could not open JPG fifo\n");
        return NULL;
    }
    while (!G_stop_rq) {
        rc = S817_BlockDone(board, 200, &aChan, &capType);
        if (rc != 0) {
            continue;
        } 
        // grab the buffer(s)
        if (capType & CAP_BMP) {
            b->bmp_frame_count[aChan - 1]++;
            ecode = S817_GetBuffer(board, aChan, &cBuf, CAP_BMP);
            fe.cBuf = cBuf;
            fe.board = board;
            rc = write(fifo_fd1, &fe, sizeof(struct fifo_element));
            if (rc <= 0) {
                printf("failed to save fifo\n");
            }
            b->frame_stats_bmp[aChan - 1]++;

        }
        if (capType & CAP_JPG) {
            b->jpg_frame_count[aChan - 1]++;
            ecode = S817_GetBuffer(board, aChan, &cBuf, CAP_JPG);
            fe.cBuf = cBuf;
            fe.board = board;
            rc = write(fifo_fd2, &fe, sizeof(struct fifo_element));
            if (rc <= 0) {
                printf("failed to save fifo\n");
            }
            b->frame_stats_jpg[aChan - 1]++;
        }

        t = current_time();
        // begin calculate display any timing information
        if (tRot0 < 0.0)
            tRot0 = t;
        dt = t - tRot0;
        tRot0 = t;
        if (tRate0 < 0.0)
            tRate0 = t;
                
        if (t - tRate0 >= 3.0) {
            int c;
            double seconds = t - tRate0;
            querySysStatus(b);
            if (b->print_stats) {
                print_stats(b, seconds);
                tRate0 = t;
                /* reset all the frames */
                for (c = 0; c < SYS_CHAN_PER_BOARD; c++) {
                    b->frame_stats_bmp[c] = 0;
                    b->frame_stats_jpg[c] = 0;
                                    
                }
            }
            tRate0 = t;
        }
    }
    /* finish up */
    b->smode = MODE_IDLE;
    close(fifo_fd1);
    close(fifo_fd2);
    printf( "Stream stopped.\n");
    prompt();
    return NULL;
}

int print_help()
{
    printf("****************************************************\n");
    printf("* Help: This application is for testing only.      *\n");
    printf("* For fuller functionality and board               *\n");
    printf("* control, please try one of the othr demo programs*\n");
    printf("* This program will start up to three boards with  *\n");
    printf("* 16 channels,                                     *\n");
    printf("* and save all JPEGs concatenated to a single file *\n");
    printf("* called jpeg_flat.                                *\n");
    printf("****************************************************\n");
    return 0;
}

char jdec_str[10];
int show_jdec_value(int jdec)
{

    switch (jdec) {
    case 0:
        printf("30 fps JPG\n");
        strcpy(jdec_str, "30");
        break;
    case 1:
        printf("24 fps JPG\n");
        strcpy(jdec_str, "24");
        break;
    case 2:
        printf("20 fps JPG\n");
        strcpy(jdec_str, "20");
        break;
    case 3:
        printf("15 fps JPG\n");
        strcpy(jdec_str, "15");
        break;
    case 4:
        printf("12 fps JPG\n");
        strcpy(jdec_str, "12");
        break;
    case 5:
        printf("10 fps JPG\n");
        strcpy(jdec_str, "10");
        break;
    case 6:
        printf("6 fps JPG\n");
        strcpy(jdec_str, "6");
        break;
    case 7:
        printf("3 fps JPG\n");
        strcpy(jdec_str, "3");
        break;
    case 8:
        printf("2 fps JPG\n");
        strcpy(jdec_str, "2");
        break;
    default:
        return -1;
    }
    return 0;

}


/*
 *Init the system and execute user commands until termination.
 */
int main( int argc, char *argv[] )
{
    int rtnval;


    if (argc < 3) {
        printf("usage s817cmd2 jdec_value diskmode(0-fopen,1-open)\n");
        return -1;
    }

    jdec = atoi(argv[1]);
    disk_mode = atoi(argv[2]);
    if (disk_mode)
        printf("using \"open\" and \"write\" for disk IO\n");
    else
        printf("using \"fopen\" and \"fwrite\" for disk IO\n");

    if (show_jdec_value(jdec) != 0) {
        printf("invalid jdec value\n");
        return -1;
    }


    print_help();
    show_jdec_value(jdec);
    
    /* Attempt to open and initialize all boards. */
    rtnval = app_init();

    // If all boards successfully opened and initialized ...
    if (rtnval == 0 ) {
        if (G_numboards >= 1) {
            exec_command( "board", "1",  "" );
            exec_command( "start", "-1", "3");
            exec_command( "jdec", "-1",  argv[1] );
            exec_command( "bdec", "-1",  "7" );
            exec_command( "jload", "-1",  "" );
        }
        if (G_numboards >= 2) {
            exec_command( "board", "2",  "" );
            exec_command( "start", "-1", "3" );
            exec_command( "jdec", "-1", argv[1] );
            exec_command( "bdec", "-1",  "7" );
            exec_command( "jload", "-1", "" );
        }
        if (G_numboards >= 3) {
            exec_command( "board", "3", "" );
            exec_command( "start", "-1", "3" );
            exec_command( "jdec", "-1", argv[1]);
            exec_command( "bdec", "-1", "7" );
            exec_command( "jload", "-1", "" );
        }
        // Execute user commands until Application Quit command is invoked.
        show_jdec_value(jdec);
        cmd_processor();
    }
                
    // Close all opened boards.
    app_shutdown();

    // Exit application.
    return 0;
}

/*
 * Open and initialize one board.  Returns: 0=success, -1=error.
 */
static int board_open(int board_num)
{
    struct BOARD *b = &bd[board_num-1];
    MODE defmode = DEFAULT_MODE;
    int aChan;
    ECODE ecode;
    SYS_STATUS st;
    /* open all channels on the board. This is done
     * at the start of the application for all channels
     * we want to use later.  If you are using less channels,
     * you may want to open fewer channels to speed things up
     * OpenChannel takes some time, but only needs done once.
     */

    b->smode = MODE_IDLE;
    b->fname[0] = 0;
    printf("board open %d\n", board_num);
    b->board = board_num;
    for (aChan = 1; aChan < 17; aChan++) {
        printf("opening channel %d\n", aChan);
        b->mode[aChan - 1] = defmode;
        ecode = S817_OpenChannel(board_num, aChan, &b->mode[aChan - 1]);
        if (ecode != 0) {
            if ((board_num != 1 && aChan != 1) || (board_num == 1)) {
                printf ("Chan open error %d; aChan %d\n", ecode, aChan);
                getPrintErrorInfo(b);
            }
            return ecode;
        } else {
            ecode = S817_SetVideoMode(board_num, aChan, &b->mode[aChan - 1].vMode);
            if (ecode != 0) {
                printf ("SetMode error %d chan %d\n", ecode, aChan);
                getPrintErrorInfo(b);
                return ecode;
            }
        }
        if ((aChan % 4) == 1)
            print_help();
    }
    printf("done opening channels.\n");
    S817_SetDateTime(board_num);

    S817_GetStatusInfo(board_num, 0, &st);
    G_fwver = st.fwVer;
    return ecode;
}


///////////////////////////////////////////////////////////////////////
// Open and initialize all boards.  Returns: -1=error, 0=success.

static int app_init( )
{
    int i;
    for (i = 1; i < MAX_BOARDS; i++) {
        printf("opening board number %d\n", i);
        if (board_open(i) != 0 ) {
            break;
        }
    }

    G_numboards = i - 1;
    printf("opened %d boards\n", G_numboards);
    if (start_threads() != 0)
        return -1;
    return (G_numboards > 0) ? 0 : -1;
}

//////////////////////////////////////////////////////////////
// Terminate all stream threads and close all boards.

static void app_shutdown(void)
{
    struct BOARD *b;
    int i;
    int j;
    // For each board ...

    stop_threads();
    for (i=0;i<G_numboards;i++) {
        // Terminate the board's I/O thread if it exists.
        for(j = 1; j <= MAX_CHANNELS; j++) {
            b = &bd[i];
            stop_stream( b, j);
        }
        // Close the board.
        printf("closing board %d\n", i + 1);
        for (j = 1; j <= MAX_CHANNELS; j++) {
            if (S817_CloseChannel(i + 1, j) != 0) {
                printf("failed to close channel\n");
            }
        }
    }
    printf( "\n" );
}

/* display the prompt */
static void prompt( void )
{
    printf( "s817demo[%d]>", act_bdnum + 1);
    fflush( stdout );
}

/* Display the command menu. */
static void show_menu( void )
{
    printf( "+-S817cmd2-COMMANDS------------------------------------------------------------+\n" );
    printf( "|   p n ........... print stats (1=on, 0 off)                                  |\n" );
    printf( "|   h[elp] ........ display this list                                          |\n" );
    printf( "|   q[uit] ........ quit program                                               |\n" );
    printf( "|   err ........... print extended debug information                           |\n" );
    printf( "|   verbose........ verbose debug                                              |\n" );
    printf( "|   board n........ switch to board n (for displaying stats)                   |\n" );
    printf( "+------------------------------------------------------------------------------+\n" );
    printf("fw version: %d.%d\n", G_fwver >> 16, G_fwver & 0xff);
    printf("saving data to file\n");

}



/* Decode and execute user command.  Returns TRUE if Quit selected. */

static int exec_command( char *cmd, char *arg,char *arg2 )
{
    struct BOARD    *b;

    // Convert argument to integer value (not relevant for all commands).
    int arg_val = -1;
    sscanf( arg, "%d", &arg_val );

    // Terminate application.
    if (cmd[0] == 'q' )
        return TRUE;

    if (strcmp( cmd, "board" ) == 0 ) {
        if (( arg_val > 0 ) && ( arg_val <= G_numboards ) ) {
            act_bdnum = arg_val - 1;
        }
        else {
            printf( "Illegal or missing argument; must be in range 1..%d\n", G_numboards);
            prompt();
            return FALSE;
        }
        prompt();
        return FALSE;
    }

    // Show help.
    if ((cmd[0] == 'h') || (cmd[0] == '?')) {
        show_menu();
        prompt();
        return FALSE;
    }
    // Prompt for user command.
    prompt();
    // Set pointer to active board's object.
    b = &bd[act_bdnum];
    if (strcmp(cmd, "get") == 0 ) {
        get_attributes(b);
    } else if (strcmp( cmd, "start"       ) == 0 ) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        if ((chn < 1 || chn > MAX_CHANNELS) && (chn !=-1)) {
            printf("invalid channel\n");
            prompt();
            return FALSE;
        }
        if (arg2[0] == 0) {
            val = CAP_JPG;
        }
        start_stream( b, chn, val);
        printf("type \"p\" followed by \"Enter\" for stats\n");
        prompt();
    } else if (strcmp( cmd, "stop"      ) == 0 ) {
        int chn,val;
        sscanf( arg, "%d", &chn);
        sscanf( arg2, "%d", &val);
        if ((chn < 1 || chn > MAX_CHANNELS) && (chn !=-1)) {
            printf("invalid channel\n");
            prompt();
            return FALSE;
        }
        stop_stream( b, (arg[0] == 0) ? 1 : arg_val);
        G_jpeg_load = 0;
        memset(G_jpegs_saved, 0, sizeof(int) * MAX_CHANNELS);
        prompt();
    } else if (strcmp( cmd, "ppm") == 0 ) {
        int chn;
        sscanf( arg, "%d",&chn);
        if (chn < 1 || chn > MAX_CHANNELS)
            chn = 1;
        strcpy(b->fname, ( arg2[0] == 0 ) ? "snapshot.ppm" : arg2);
        if ((b->mode[chn-1].cMode.capType & CAP_BMP) == 0) {
            printf("stream must be running with BMP capture!\n");
            prompt();
        }
        else {
            // fire snapshot in IO thread
            b->snapshot_ppm[chn-1] = 1;
        }
    } else if ((strcmp( cmd, "jpg") == 0) || (strcmp( cmd, "jpeg") == 0)) {
        int chn;
        sscanf( arg, "%d",&chn);
        if (chn < 1 || chn > MAX_CHANNELS)
            chn = 1;
        strcpy(b->fname, ( arg2[0] == 0 ) ? "snapshot.jpg" : arg2);
        if ((b->mode[chn-1].cMode.capType & CAP_JPG) == 0) {
            printf("stream must be running with JPG capture!\n");
            prompt();
        }
        else {
            // fire snapshot in IO thread
            b->snapshot_jpeg[chn-1] = 1;
        }
    } else if (strcmp( cmd, "p") == 0 ) {
        int val;
        sscanf( arg, "%d",&val);
        b->print_stats = val;
    } else if (strcmp( cmd, "jload") == 0 ) {
        int chn;
        sscanf( arg, "%d",&chn);
        if ((chn < 1 || chn > MAX_CHANNELS) && (chn !=-1)) {
            printf("invalid channel\n");
            prompt();
            return FALSE;
        }
        G_jpeg_load = 1;
    } else if (strcmp( cmd, "bdec" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (((chn > MAX_CHANNELS || chn < 1) && (chn != -1)) || val < 0 || val > 8) {
            printf("bad channel or decimation value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            if (chn == -1 ) {
                for (chn=1;chn<=16;chn++) {
                    b->mode[chn-1].cMode.frameRateBmp = val;
                    ret = S817_SetCaptureMode( b->board, chn, &b->mode[chn-1].cMode);
                    if (ret != 0 ) {
                        printf("failed set mode 0x%x\n", ret);
                        prompt();
                    }
                }
            } else {
                b->mode[chn-1].cMode.frameRateBmp = val;
                ret = S817_SetCaptureMode( b->board, chn, &b->mode[chn-1].cMode);
                if (ret != 0 ) {
                    printf("failed set mode 0x%x\n", ret);
                    prompt();
                }
            }
        }
    } else if (strcmp( cmd, "jdec" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (((chn > MAX_CHANNELS || chn < 1) && (chn != -1)) || val < 0 || val > 8) {
            printf("bad channel or decimation value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            if (chn == -1 ) {
                for (chn=1;chn<=16;chn++) {
                    b->mode[chn-1].cMode.frameRateJpg = val;
                    ret = S817_SetCaptureMode( b->board, chn, &b->mode[chn-1].cMode);
                    if (ret != 0 ) {
                        printf("failed set mode 0x%x\n", ret);
                        prompt();
                    }
                }
            } else {
                b->mode[chn-1].cMode.frameRateJpg = val;
                ret = S817_SetCaptureMode( b->board, chn, &b->mode[chn-1].cMode);
                if (ret != 0 ) {
                    printf("failed set mode 0x%x\n", ret);
                    prompt();
                }
            }
        }
    } else if (strcmp( cmd, "scale" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > MAX_CHANNELS || chn < 1 || val <0 || val > 3) {
            printf("bad channel or scale value %d %d\n",chn,val);
            prompt();
        } else {
            int ret;
            b->mode[chn-1].vMode.jpgSize = val;
            ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "bright" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > MAX_CHANNELS || chn < 1 || val <0 || val > 255) {
            printf("bad channel or bright value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].vMode.brightness = val;
            ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "contrast" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > MAX_CHANNELS || chn < 1 || val <0 || val > 255) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].vMode.contrast = val;
            ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "vsat" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > MAX_CHANNELS || chn < 1 || val <0 || val > 255) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].vMode.saturation = val;
            ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "vhue" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > MAX_CHANNELS || chn < 1 || val <0 || val > 255) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].vMode.hue = val;
            ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "jqual" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > MAX_CHANNELS || chn < 1 || val <0 || val > 100) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].cMode.compQuality = val;
            ret = S817_SetCaptureMode( b->board, chn, &b->mode[chn-1].cMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "osd" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > MAX_CHANNELS || chn < 1 || val <0 || val > 1) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        } else {
            OSD_MODE osd;
            int ret;
            memset(&osd, 0, sizeof(osd));
            osd.osdOn = val;
            osd.osdBmp = 1;
            osd.transparent = 1;
            osd.positionTop = 1;
            osd.ddmm = 1;
            osd.year2 = 0;
            osd.fraction = 2;
            strncpy(osd.line, "Date: ^d Time: ^t", 80);
            ret = S817_SetOsdMode(b->board, chn, &osd);
            if (ret != 0 ) {
                printf("failed set OSD mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "vsystem" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > MAX_CHANNELS || chn < 1 || val < 0 || val >= TV_MAX) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        } else {
            int ret;
            b->mode[chn-1].vMode.tvFormat = val;
            if (b->mode[chn-1].cMode.capType != CAP_OFF) {
                printf("You must stop the stream first before changing the video system\n");
                prompt();
            } else {
                ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
                if (ret != 0 ) {
                    printf("failed set video system\n");
                    prompt();
                } else {
                    printf("changed to %s on channel %d\n", val == 0 ? "NTSC" : "PAL",
                           chn);
                }
            }
        }
    } else if (strcmp(cmd, "err") == 0) {
        getPrintErrorInfo(b);
        prompt();
    } else if (strcmp(cmd, "verbose") == 0) {
        printf("verbose debugging on\n");
        G_verbose = 1;
        prompt();
    } else if (cmd[0] != 0 ) {
        printf( "Invalid command: %s\n", cmd );
        prompt();
    }

    return FALSE;
}

////////////////////////////////////////////////////////////////////
// Execute user commands until Application Quit is requested.

static void cmd_processor( void )
{
    char                    cmdbuf[MAXLEN];
    char                    cmd[MAXLEN];
    char                    arg[MAXLEN];
    char                    arg2[MAXLEN];
    char                    *c;
    // Display the command menu.
    show_menu();

    // Prompt for user command.
    prompt();

    // Repeat until Application Terminate is requested ...
    for ( ; ; )
    {
        // Input user command string, and if no errors encountered ...
        if (fgets( cmdbuf, MAXLEN - 1, stdin ) != NULL )
        {
            // Truncate newline char at end of command string.
            cmdbuf[ strlen( cmdbuf ) - 1 ] = 0;

            // If command line is not blank ...
            if (cmdbuf[0] != 0 )
            {
                // Force command string to lower case, except command=="text".
                if (strncmp( cmdbuf, "text", 4 ) != 0 )
                    for ( c = cmdbuf; *c != 0; c++ ) 
                        *c = tolower( *c );

                // Parse command string into opcode and argument.
                cmd[0] = arg[0] = 0;
                arg2[0] = 0;
                sscanf( cmdbuf, "%s %s %s", cmd, arg,arg2 );
                // Decode and execute command
                if (exec_command( cmd, arg,arg2 ) )
                    break;
            }
            else
            {
                // Prompt for user command.
                prompt();
            }
        }
    }
}



// Fetch and display the attributes.
static void get_attributes( struct BOARD *b )
{
    int i;
    char status[100];
    char tmp[100];
    int bPAL;
    int fr_bmp;
    int fr_jpg;
    char *fr_ntsc[9] = { "30fps", "24fps", "20fps", "15fps", "12fps",
                         "10fps", "6fps", "3fps", "2fps"};
    char *fr_pal[9] = { "25fps", "20fps", "16.7fps", "12.5fps", "10fps",
                        "8.3fps", "5fps", "2.5fps", "1.7fps"};

    printf( "\nCURRENT SETTINGS -----------\n" );
    printf( "\nBoard %d----------\n", b->board);
    for (i=0;i<MAX_CHANNELS;i++) {
        bPAL = b->mode[i].vMode.tvFormat == TV_PAL ? 1 : 0;
        fr_bmp = b->mode[i].cMode.frameRateBmp;
        fr_jpg = b->mode[i].cMode.frameRateJpg;
        switch (b->mode[i].cMode.capType) {
        case CAP_BMP:
            sprintf(status, "BMP(%s)", bPAL ? fr_pal[fr_bmp] : fr_ntsc[fr_bmp]);
            break;
        case CAP_JPG:
            sprintf(status, "JPG(%s)", bPAL ? fr_pal[fr_bmp] : fr_ntsc[fr_bmp]);
            break;
        case (CAP_JPG | CAP_BMP):
            sprintf(status, "BMP(%s)+JPG(%s)",  bPAL ? fr_pal[fr_bmp] : fr_ntsc[fr_bmp],
                    bPAL ? fr_pal[fr_jpg] : fr_ntsc[fr_jpg]);
            break;
        case CAP_OFF:
        default:
            sprintf(status, "OFF");
            break;
        }
        snprintf(tmp, 100, "channel %d, capture %s", i, status);
        puts(tmp);
    }
}


static int start_threads(void)
{
    int rc;
    pthread_attr_t attr;
    pthread_attr_init( &attr );
    pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );

    if (G_smode != MODE_IDLE ) {
        // Abort command.
        printf( "already running\n" );
        prompt();
        return 0;
    }
    if (access(FIFO_NAME_BMP, F_OK) == -1) {
        rc = mkfifo(FIFO_NAME_BMP, 777);
        if (rc != 0) {
            printf("could not create fifo\n");
            return -ENOMEM;
        }
    }
    if (access(FIFO_NAME_JPG, F_OK) == -1) {

        rc = mkfifo(FIFO_NAME_JPG, 777);
        if (rc != 0) {
            printf("could not create fifo\n");
            return -ENOMEM;
        }
    }
    G_smode = MODE_STREAM;
    prompt();
    pthread_create( &G_io_thread_idJ, NULL, jpg_thread_func, NULL );
    pthread_create( &G_io_thread_idB, NULL, bmp_thread_func, NULL);
    pthread_attr_destroy( &attr );
    return 0;
}

static void stop_threads(void)
{
    if (G_smode != MODE_IDLE ) {
        // Request thread termination.
        G_stop_rq = 1;
        // Wait for the threads to terminate
        pthread_join(G_io_thread_idJ, NULL);
        pthread_join(G_io_thread_idB, NULL);
    }
    return;
}


static void start_stream_inner( struct BOARD *b, int chan, int mask)
{
    MODE *pmode = &b->mode[chan-1];
    ECODE ecode;
    char *mode_str[] = {"JPG", "BMP", "JPG+BMP" };
    /* start capture for this channel */


    pmode->cMode.capMode = CAP_MODE_CONT;
    pmode->cMode.capType = mask;
        
    /* board, channel, &cMode */
    ecode = S817_SetCaptureMode(b->board, chan, &pmode->cMode);
    if (ecode) {
        printf("Capture error %d; channel %d board %d\n", ecode, chan, b->board);
        print_ecode(ecode);
        return;
    }
        
    printf("starting channel %d in %s mode\n", chan, mode_str[pmode->cMode.capType-1]);
    b->b_stat_exclude[chan-1] = 1;
    b->smode = MODE_STREAM;
    return;
}

static void start_stream( struct BOARD *b, int chan, int mask )
{
    int i;
    pthread_create( &b->io_thread_idM, NULL, main_thread_func, b);
    i = chan;
    if (chan <= 0 || chan > SYS_CHAN_PER_BOARD) {
        printf("starting all channels\n");
        prompt();
        for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
            start_stream_inner(b, i+1, mask);
        }
                
    } else {
        start_stream_inner(b, chan, mask);
    }
    return;
}

//////////////////////////////////////////////////
// Halt data stream

static void stop_stream_inner( struct BOARD *b, int chan )
{
    MODE *pmode = &b->mode[chan-1];
    ECODE ecode;
    pmode->cMode.capType = CAP_OFF;
    ecode = S817_SetCaptureMode(b->board, chan, &pmode->cMode);
    if (ecode) {
        printf("Capture error %d; channel %d\n", ecode, chan);
        prompt();
        return;
    }
    b->stop_rq = 1;
    pthread_join(b->io_thread_idM, NULL);

    b->smode = MODE_STREAM;
        
}
        

static void stop_stream( struct BOARD *b, int chan )
{
    if (chan == -1) {
        int i;
        printf("stopping all channels\n");
        prompt();
        for (i = 0; i < MAX_CHANNELS; i++)
            stop_stream_inner(b, i+1);
    } else
        stop_stream_inner(b, chan);
    return;
}

/* return current time (in seconds) */
static double current_time(void)
{
#if defined(OSTYPE_LINUX)
    struct timeval tv;
    struct timezone tz;
    (void) gettimeofday(&tv, &tz);
    return (double) tv.tv_sec + tv.tv_usec / 1000000.0;
#else
    static BOOL                         first_time = TRUE;      
    LARGE_INTEGER                       cur_time;
    static LARGE_INTEGER        freq;
    static BOOL                         has_timer = FALSE;
    QueryPerformanceCounter(&cur_time);

    if ( first_time )
    {
        first_time = FALSE;
        has_timer = QueryPerformanceFrequency(&freq);
    }
    return (double)cur_time.QuadPart / freq.QuadPart;
#endif
}



static void print_stats(BOARD *b, double seconds)
{
    int i;

    if (seconds == 0)
        return;
    printf("\n====================================\n");
    printf("current stats: (Enter \"p 0\" to turn off stats)\n");
    printf("last update %lf seconds (type h for help)\n", seconds);
    for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
        double fps_jpg = b->frame_stats_jpg[i] / seconds;
        double fps_bmp = b->frame_stats_bmp[i] / seconds;

        if (b->mode[i].cMode.capType == CAP_OFF)
            continue;

        if (b->b_stat_exclude[i]) {

            printf("chan %d: %d JPG frames(just started, no fps calculation)\n", i + 1,
                   b->frame_stats_jpg[i]);
            b->b_stat_exclude[i] = 0;
                        
        } else {
            printf("chan %d: %d(bmp) %d(jpg) frames, %6.3f fps(bmp) %6.3f fps(jpg)", i + 1,
                   b->frame_stats_bmp[i],
                   b->frame_stats_jpg[i],
                   fps_bmp,fps_jpg);
            if (G_verbose)
                printf(" rsrvd[1]: 0x%02x", G_rsrv[i]);
            printf("\n");
        }
    }
    printf("====================================\n");
    prompt();
    return;
}

/* error codes to text */
static void print_ecode(ECODE ecode)
{
    switch (ecode) {
    case ERR_SX_OPEN:
        printf("unable to open the driver\n");
        break;
    case ERR_BRIDGE_NOT_FOUND:
        printf("no 817 bridge found\n");
        break;
    case ERR_SELECT_DEVICE:
        printf("selected board not found\n");
        break;
    case ERR_MEM_ALLOC:
        printf("memory allocation error\n");
        break;
    case ERR_MEM_LOCK:
        printf("DMA buffer lock error\n");
        break;
    case ERR_I2CINIT_BUSY:
        printf("I2C bus initialization error\n");
        break;
    case ERR_I2CINIT_ERR:
        printf("I2C bus initialization error\n");
        break;
    case ERR_TIMEOUT:
        printf("internal access timeout\n");
        break;
    case ERR_LOADING_DLL:
        printf("error loading DLL\n");
    case ERR_FUNCADDR:
        printf("error getting a DLL func\n");
        break;
    case ERR_DLL:
        printf("error accessing DLL\n");
        break;
    case ERR_NO_REGISTERED_DEV:
        printf("unable to register boards\n");
        break;
    case ERR_I2C:
        printf("internal I2C bus error\n");
        break;
    case ERR_I2CTO:
        printf("internal I2C timeout\n");
        break;
    case ERR_EEPROM:
        printf("EEPROM programming error\n");
        break;
    case ERR_XPTO:
        printf("internal crosspoint switch timeout\n");
        break;
    case ERR_ILLEGAL_MODE:
        printf("illegal value in MODE structure\n");
        break;
    case ERR_INTERNAL_OBJ:
        printf("internal obj allocation error\n");
        break;
    case ERR_ACQ_TIMEOUT:
        printf("acquisition timeout\n");
        break;
    case ERR_INTERRUPT:
        printf("interrupt init error\n");
        break;
    case ERR_ILLEGAL_INDEX:
        printf("illegal board index\n");
        break;
    case ERR_PCI:
        printf("PCI access error\n");
        break;
    case ERR_BOARDADDR:
        printf("board address mapping failed\n");
        break;
    case ERR_BOARDREG:
        printf("error registering the board\n");
        break;
    case ERR_ITEMS:
        printf("board has no items\n");
        break;
    case ERR_ALREADYOPEN:
        printf("board already open\n");
        break;
    case ERR_FIFOEMPTY:   
        printf("data FIFO empty\n");
        break;
    case ERR_DATA_ACCESS_TO:
        printf("internal process data access timeout\n");
        break;
    case ERR_DEVICEINFO:
        printf("driver error\n");
        break;
    case ERR_DSP_INIT:
        printf("dsp init\n");
        break;
    case ERR_BOARD_NOT_FOUND:
        printf("board not found\n");
        break;
    case ERR_ILLEGAL_BOARD:
        printf("illegal board index\n");
        break;
    case ERR_ILLEGAL_CHAN:
        printf("illegal channel #\n");
        break;
    case ERR_NO_FREE_BUFFER:
        printf("no free buffers\n");
        break;
    case ERR_DSP_START_TO:
        printf("DSP start timeout\n");
        break;
    case ERR_DSP_CMD_TO:
        printf("HOST-to-DSP command response timeout 40\n");
        break;
    case ERR_DSP_CMD_BUSY:
        printf("HOST-to-DSP command send error (DSP busy)\n");
        break;
    case ERR_DSP_DATA_TO:
        printf("DSP timeout while receiving host data\n");
        break;
    case ERR_DSP_PARSE_DATA:
        printf("DSP parse error\n");
        break;
    case ERR_ALLOC_FIFO:
        printf("error allocating FIFO\n");
        break;
    case ERR_FIFO_EMPTY:
        printf("FIFO contains no data\n");
        break;
    case ERR_DSP_INVALID_ADDR:
        printf("invalid address passed to DSP\n");
        break;
    case ERR_DSP_DATA_TO_CAP:
        printf("data timeout cap\n");
        break;
    case ERR_ALREADY_CAPTURING:
        printf("channel already capturing (StartCapture attempted)\n");
        break;
    case ERR_NOT_CAPTURING:
        printf("channel not capturing (StopCapture attempted)\n");
        break;
    case ERR_ILLEGAL_COMPQUALITY:
        printf("illegal compression quality value 50\n");
        break;
    case ERR_ILLEGAL_FRAMERATE:
        printf("illegal frame rate\n");
        break;
    case ERR_ILLEGAL_TVFORMAT:
        printf("illegal TV format\n");
        break;
    case ERR_ILLEGAL_JPGSIZE:
        printf("illegal jpeg size\n");
        break;
    case ERR_FW_FILE_OPEN:
        printf("error opening firmware file\n");
        break;
    case ERR_FW_FILE_READ:
        printf("error reading firmware file\n");
        break;
    case ERR_ILLEGAL_DSP_INDEX:
        printf("illegal dsp index\n");
        break;
    case ERR_ILLEGAL_BUF:
        printf("illegal buffer index\n");
        break;
    case ERR_ADDR_NOT_CONFIG:
        printf("DSP address not configured on host\n");
        break;
    case ERR_ILLEGAL_CAPMODE:
        printf("illegal capture mode\n");
        break;
    }
    return;
}

/* error info */
static void getPrintErrorInfo(struct BOARD *b)
{
    int j, k;
    int errors;
    int dsp;
    EXTERR extErr[SYS_EXTERR_BUF_SIZE];
        
    for (dsp=0; dsp< 4; dsp++) {
        for (j = 0; j < 4; j++) {
            if (S817_GetExtErrInfo(b->board, dsp, j, extErr) != 0) {
                return;
            }
            errors = extErr[0].tick;
            if (errors) {
                printf("DSP %d chan %d Errors %d\n", dsp, j, errors);
                errors = (errors <= SYS_EXTERR_BUF_SIZE - 1) ? errors : SYS_EXTERR_BUF_SIZE - 1;
                for (k = 0; k < errors; k++) {
                    printf ("%08X  %04d  %08X  %08X\n", extErr[k + 1].tick, extErr[k + 1].ecode,
                            extErr[k + 1].reserved[0], extErr[k + 1].reserved[1]);
                }
            }
        }
        /* PCI thread errors */
        if (S817_GetExtErrInfo(b->board, dsp, 4, extErr) != 0) {
            return;
        }
        errors = extErr[0].tick;
        if (errors) {
            printf("DSP %d chan PCI; Errors %d\n", dsp, errors);
            errors = (errors <= SYS_EXTERR_BUF_SIZE - 1) ? errors : SYS_EXTERR_BUF_SIZE - 1;
            for (k = 0; k < errors; k++) {
                printf ("%08X  %04d  %08X  %08X\n", extErr[k + 1].tick, extErr[k + 1].ecode,
                        extErr[k + 1].reserved[0], extErr[k + 1].reserved[1]);
            }
        } else {
            printf("exterrinfo: no errors found\n");
        }
    }
    getSysStatus(b);

    return;
}


static void getSysStatus(struct BOARD *b)
{
    int jj;
    int dsp;
    SYS_STATUS st;
    for (dsp=0; dsp< 4; dsp++) {
        S817_GetStatusInfo(b->board, dsp, &st);
        printf("sysstat dsp %d\n", dsp);
        for (jj=0;jj<4;jj++) {
            printf("[%d] sysStat lastCmd %d ticks %d reserved: 0x%02x 0x%02x 0x%02x 0x%02x\n",
                   jj, st.chanData[jj].lastCmd, 
                   st.chanData[jj].chanTick,
                   st.reserved[0],
                   st.reserved[1],
                   st.reserved[2],
                   st.reserved[3]);
            printf("[%d] sysStat lockPend %d putToSleep %d fVidStat %d noFreeJPG %d noFreeBmp %d\n",
                   jj, st.chanData[jj].lockPend,
                   st.chanData[jj].putToSleep,
                   st.chanData[jj].fvidStat,
                   st.chanData[jj].noFreeBufJpg,
                   st.chanData[jj].noFreeBufBmp);
        } 
        printf("sys stat fw %d.%d ",
               (st.fwVer >> 16), (st.fwVer & 0xff));
        printf("tot irq %d\n",
               st.reserved[0]);
    }
    return;
}


static void querySysStatus(struct BOARD *b)
{
    int jj;
    int dsp;
    SYS_STATUS st;
    static int noJpg[SYS_BOARDS][MAX_CHANNELS];
    static int noBmp[SYS_BOARDS][MAX_CHANNELS];
    static int calls[SYS_BOARDS] = {0,0,0,0};
    int bidx = b->board - 1;
    int bChan;
    int no_print = 1;
    for (dsp=0; dsp< 4; dsp++) {
        S817_GetStatusInfo(b->board, dsp, &st);
        for (jj=0;jj<4;jj++) {
            bChan = dsp * 4 + jj;
            if (calls[bidx] > 2) {
                if (noJpg[bidx][bChan] != st.chanData[jj].noFreeBufJpg) {
                    printf("dropping %d JPGs on board %d channel %d\n",
                           st.chanData[jj].noFreeBufJpg - noJpg[bidx][bChan],
                           bidx+1, bChan +1);
                    no_print = 0;
                }
                if (noBmp[bidx][bChan] != st.chanData[jj].noFreeBufBmp) {
                    printf("dropping %d BMPs on board %d channel %d\n",
                           st.chanData[jj].noFreeBufBmp - noBmp[bidx][bChan],
                           bidx+1, bChan +1);
                    no_print = 0;
                }
            }
            noJpg[bidx][bChan] = st.chanData[jj].noFreeBufJpg;
            noBmp[bidx][bChan] = st.chanData[jj].noFreeBufBmp;
        }
    }
    if (no_print) {
        printf("saving data(%s fps JPG). no dropped frames\n", jdec_str);
        prompt();
    }
    calls[b->board-1]++;
    return;
}


